#!/usr/bin/perl

# Description
# ===========
# This script backs up MySQL databases to file.
# See readme.txt for more information.
########### Script start ############

## Dont touch this section. See the USER DEFINED VARIABLES section for stuff to modify when deploying this script.

# Use the following pragmas to help debugging and keep sane
use strict;
use warnings;

## Include some packages we need.

# Include the ability to get file properties.
use File::stat;

# Include the ability to get the current directory of this script.
use Cwd 'abs_path';

# Include the ability to compress files.
use Archive::Zip qw( :ERROR_CODES :CONSTANTS );

# Include the ability to list files in a directory using wildcards.
use File::DosGlob 'glob';

# Include the fileparse function and others.
use File::Basename;

# Include the ability to send email via SMTP.
use Net::SMTP;

# Include the ability to get the hostname (the hostname function).
use Sys::Hostname;

#######################################################################################################
########### USER DEFINED VARIABLES (change these to suit your requirements and environment) ###########
#######################################################################################################

# Backup directory ( where the backup files are stored ).
my $backup_dir = '/home/apps/backups/mysql_backups';

# Path the mysqldump executable (not to it's folder).
my $path_to_mysqldump = '/usr/bin/mysqldump';
# Path the mysqladmin executable (not to it's folder).
my $path_to_mysqladmin = '/usr/bin/mysqladmin';

# Database information array (a 2D array).
# Element description: host, db, username, password.
# Note: Each element within the @databases array is itself an array (of strings).
# There should not be a comma at the end of the last line but should be at end of other lines if 
# backing up more than 1 database.
my @databases = (
	# Database 1
	['127.0.0.1', 'mysql', 'root', '12345678']
	# Database 2
	#['127.0.0.1', 'piwik', 'root', '12345678']
	# Database 3
	#['192.168.0.2', 'database3', 'username', 'password']
);

# The script error log filename
# Script directory ( current directory of this script ).
use File::Basename;
my $scriptdir = dirname(__FILE__);
my $error_log = "$scriptdir/mysql_backup.log";

# The mysqldump warning and error log. If this file contains anything then there was a problem with the backup job
# and it will be used by the script. It will be overwritten with each execution of the script so is not meant to be
# for human reference.
my $mysqldump_error_log = "$scriptdir/mysqldump_error.log";

# dump all databases into a single file rather than create individual files for each database
my $all_databases = 1;

# The mysqldump options to use (dont change unless you know what these mean). Dont include --all-databases as thats inserted automatically by the script if you set the $all_databases variable to 1.
my $mysqldump_options = '--events --routines --single-transaction --flush-logs --master-data=2 --log-error=' . $mysqldump_error_log;

# Whether to compress the backup files afterwards with zlib compression or not (1 = yes, 0 = no). 
# The files are readable with the freeware app "7-zip" but not by the built-in Windows 7 version of zip.
my $use_compression = 1;

# Threshold in days that old backup files (.zip only) should be deleted, set to 0 to disable deletion of old backup files.
my $delete_files_older_than_days = 7;

# Whether to send an email to someone if the backup job fails (1 = use email, 0 = don't use email).
my $use_email_alerting = 1;

# My domain name, used by send mail function
my $domainname = 'mydomain.com';

# SMTP Server for email alerts when the backup job fails.
my $smtp_server = 'smtp.mydomain.com';

# E-mail address where the e-mail will be sent from when the backup job fails.
my $mail_from = 'support@mydomain.com';

# E-mail address where the e-mail will be sent to the backup job fails.
my $mail_to = 'support@mydomain.com';

# E-mail subject for email alerts when the backup job fails.
my $mail_subject = 'MySQL backup script/job failed on host ' . hostname;

# Debug variable. Set to 1 to print debug info to STDOUT
my $debug = 1;
########### System global variables #############


my $backup_mode;

# Declare date and time stamp variables for use with get_date_stamp subroutine.
my $log_date_stamp;
my $file_time_stamp;
my $date_stamp;
my $today;
my $time_now;
my $epoch_secs;
my $alert_detail = "";


########### Start of Main program ############

# Get date stamp.
get_date_stamp();

########### Initialise log files ##########

# Open log file with datestamp extension.
open ( ERRLOG, ">>$error_log" ) or die "$log_date_stamp - Cannot open file $error_log for writing: $!";
print ERRLOG "-------------------------------------------------------------------------------\n";

# Do initial sanity check.
initial_check();

if ($backup_mode eq "FULL") {
	print "Starting full backup of MySQL databases to file.\n" if ($debug);
	print ERRLOG "$log_date_stamp - Starting full backup of MySQL databases to file.\n";

	# Main loop.
	# For each reference to an array within the @databases array...
	foreach my $dbinfo ( @databases ) {
		get_date_stamp();
		
		# Assign values from the array into easy to read scalars. $dbinfo is a reference to an array so we 
		# need to deference it's elements to get the values.
		# Dereferencing means getting at the value that a reference points to.
		my $database_host = $$dbinfo[0];
		my $database_name = $$dbinfo[1];
		my $database_user = $$dbinfo[2];
		my $database_pass = $$dbinfo[3];
		
		my $dump_filename;
		if ( $all_databases ) {
			$dump_filename = hostname . "_" . $file_time_stamp . ".sql";
		}
		else {
			$dump_filename = $$dbinfo[1] . "_" . $file_time_stamp . ".sql";
		}
		
		my $dump_filepath = "${backup_dir}/$dump_filename";
		
		# Execute mysqldump utility.
		my $command;
		if ( $all_databases ) {
			if ( $database_pass eq '' ) {
				$command = "$path_to_mysqldump --user=$database_user --host=$database_host --all-databases $mysqldump_options > $dump_filepath";
			}
			else {
				$command = "$path_to_mysqldump --user=$database_user --password=$database_pass --host=$database_host --all-databases $mysqldump_options > $dump_filepath";
			}
		}
		else {
			if ( $database_pass eq '' ) {
				$command = "$path_to_mysqldump --user=$database_user --host=$database_host $mysqldump_options $database_name > $dump_filepath";
			}
			else {
				$command = "$path_to_mysqldump --user=$database_user --password=$database_pass --host=$database_host $mysqldump_options $database_name > $dump_filepath";
			}
		}
		if ( $all_databases ) {
			print ERRLOG "$log_date_stamp - Backing up all databases to file $dump_filepath\n";
		}
		else {
			print ERRLOG "$log_date_stamp - Backing up database: $database_name to file $dump_filepath\n";
		}
		print "Executing command: $command\n" if ($debug);
		print ERRLOG "$log_date_stamp - Executing command: $command\n";
		my @args = ($command);
		system(@args);
		
		# If there was a problem then record the details.
		if ($? != 0) 
		{
			get_date_stamp();
			open (MYSQLDUMPLOG, "$mysqldump_error_log") or print ERRLOG "$log_date_stamp - Cannot read mysqldump error log file $mysqldump_error_log\n";
			my @mysqldumperrors = <MYSQLDUMPLOG>;
			close MYSQLDUMPLOG;
			unlink $mysqldump_error_log;
			my $str = join("", @mysqldumperrors);
			
			if ( $all_databases ) {
				print "Failed to backup all databases\n" if ($debug);
				print ERRLOG "$log_date_stamp - Failed to backup all databases using the \"mysqldump\" utility. Reason according to mysqldump: \n" . $str . "\n";
				$alert_detail .= "Problem\n-------\nFailed to backup all databases using the \"mysqldump\" utility.\n\nReason\n------\n" . $str . "\n\n";
			}
			else {
				print "Failed to backup the database $database_name\n" if ($debug);
				print ERRLOG "$log_date_stamp - Failed to backup database \"$database_name\" using the \"mysqldump\" utility. Reason according to mysqldump: \n" . $str . "\n";
				$alert_detail .= "Problem\n-------\nFailed to backup database \"$database_name\" using the \"mysqldump\" utility.\n\nReason\n------\n" . $str . "\n\n";
			}
			next;
		}
		# Delete the mysqldump log file.
		unlink $mysqldump_error_log;
		
		# Compress the backup files if the option was specified.
		get_date_stamp();
		if ( $use_compression ) {
			print ERRLOG "$log_date_stamp - Compressing ${backup_dir}/$dump_filename\n";
			my $original_file_stat = stat( "$backup_dir/$dump_filename" );
			print ERRLOG "$log_date_stamp - File size of $dump_filename = " . $original_file_stat->size . " bytes\n";
			my $compressed_filename = compress_file ( "${backup_dir}/$dump_filename" );
			get_date_stamp();
			if (! defined ($compressed_filename)) {
				print ERRLOG "$log_date_stamp - File compression failed\n";
				$alert_detail .= "$log_date_stamp - File compression failed\n\n";
			}
			else {
				my $compressed_file_stat = stat( "$backup_dir/$dump_filename.zip" );
				if (! $compressed_file_stat ) {
					print ERRLOG "$log_date_stamp - Can't get properties of file ${backup_dir}/$dump_filename: $!";
				} 
				else {
					print ERRLOG "$log_date_stamp - File size of $dump_filename.zip = " . $compressed_file_stat->size . " bytes\n";
				}
			}
		}
	}

	# Delete old backup files if the option was specified.
	get_date_stamp();
	if ($delete_files_older_than_days > 0) {
		print ERRLOG "$log_date_stamp - Getting list of zip files older than $delete_files_older_than_days days\n";
		my $filelist = list_files_older_than_days($backup_dir, '.zip', $delete_files_older_than_days);
		if (! defined($filelist)) {
			print ERRLOG "$log_date_stamp - No files older than $delete_files_older_than_days days to delete\n";
		}
		else {
			print ERRLOG "$log_date_stamp - Deleting backup files that are older than $delete_files_older_than_days days\n";
			foreach my $filename (@$filelist) {
				print ERRLOG "$log_date_stamp - Deleting file $filename\n";
				unlink $filename;
			}
		}
	}
}
else {
	# flush logs for the incremental backup
	print "Starting incremental backup of MySQL databases (flushing the binary log).\n" if ($debug);
	print ERRLOG "$log_date_stamp - Starting incremental backup of MySQL databases (by running the FLUSH LOGS command).\n";

	# Main loop.
	# For each reference to an array within the @databases array...
	foreach my $dbinfo ( @databases ) {
		get_date_stamp();
		
		# Assign values from the array into easy to read scalars. $dbinfo is a reference to an array so we 
		# need to deference it's elements to get the values.
		# Dereferencing means getting at the value that a reference points to.
		my $database_host = $$dbinfo[0];
		my $database_name = $$dbinfo[1];
		my $database_user = $$dbinfo[2];
		my $database_pass = $$dbinfo[3];
		my $command;
		if ( $database_pass eq '' ) {
			$command = "$path_to_mysqladmin --user=$database_user --host=$database_host flush-logs binary";
		}
		else {
			$command = "$path_to_mysqladmin --user=$database_user --password=$database_pass --host=$database_host flush-logs binary";
		}
		print "Executing command: $command\n" if ($debug);
		print ERRLOG "$log_date_stamp - Executing command: $command\n";
		# execute the command and capture STDOUT and STDERR
		my $output = `$command 2>&1`;

		get_date_stamp();
		print "Finished executing command: $command\n" if ($debug);
		# If there was a problem then record the details.
		if ($? != 0) {
			print "Failed to execute command \"$command\"\n" if ($debug);
			print ERRLOG "$log_date_stamp - Failed to execute command \"$command\"\n";
			$alert_detail .= "Problem\n-------\nFailed to flush binary logs for database server \"$database_host\" using the \"mysqladmin\" utility.\n\nReason\n------\n" . $output . "\n\n";
		}
		else {
			print "Incremental backup completed (binary log flushed successfully)\n" if ($debug);
			print ERRLOG "Incremental backup completed (binary log flushed successfully)\n" if ($debug);
		}
	}
}
# If there was a problem backing up a database or compressing a backup file, send an email
# to the alert email address (if the email option was specified).
if ( $use_email_alerting )
{
	if ( $alert_detail ne "")
	{
		$alert_detail .= "This email was produced by the script: " . abs_path($0) . "\n";
		my $mail_body = $alert_detail;
		get_date_stamp();
		print "Sending email using following params : smtp server: $smtp_server, from: $mail_from, to: $mail_to, subject: $mail_subject, body: $mail_body\n" if ($debug);
		my $response = send_email($smtp_server, $mail_from, $mail_to, $mail_subject, $mail_body);
		if (! $response) {
			print ERRLOG "$log_date_stamp - Send email failed.\n";
			print "$log_date_stamp - Send email failed.\n";
		}
	}
}

get_date_stamp();
print "Script finished\n";
print ERRLOG "$log_date_stamp - Script finished.\n";

############# Subroutines and functions #############

### Function to check a few things before we bother going into the main section of the script
### Paramters: None
### Return value: 1 if everything is ok, a string if is something is wrong
sub initial_check {
	# check for 1 command line parameter
	$backup_mode = $ARGV[0];
	die "Backup mode not supplied as first argument to script call. Either supply FULL or INC as the value for the parameter. e.g. mysql_backup.pl FULL\n" 
		if (not defined $backup_mode);
	
	die "A backup mode of \"$backup_mode\" is invalid" if ($backup_mode ne "FULL" && $backup_mode ne "INC");
	print "Backup mode is $backup_mode\n" if ($debug);

	# check existence of backup_dir
	die "Backup directory $backup_dir does not exist! You should probably create it.\n" if (! -d $backup_dir);
	
	# check for existence of mysqldump executable
	die "The mysqldump executable at $path_to_mysqldump does not exist! Are you sure you got the path correct?\n" if (! -f $path_to_mysqldump);
	
	# if we get this far then everything is ok
}

### Subroutine to update some global variables for use in outputting a timestamp to log files
### or to get the number of seconds since 1 Jan 1970 for determining how old files are.
### Declare the $date_stamp, $file_time_stamp, $log_date_stamp and $epoch_secs as globals
### at the top of this script.
### Parameters: None
### Return value: None
sub get_date_stamp {
	#get current date and time
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	
	# correct the month and year variable values
	$mon += 1;
	$year += 1900;
	
	# Update the global timestamp vars .... declare these up top
	
	# set datestamp for logfile extension
	$date_stamp = sprintf("%02d%02d%04d", $mday, $mon, $year);
	$file_time_stamp = sprintf("%02d%02d%02d_%02d%02d%02d", $year,$mon,$mday,$hour,$min,$sec);
	# set datestamp for logfile content
	$log_date_stamp = sprintf("%02d/%02d/%04d %02d:%02d:%02d", $mday, $mon, $year, $hour, $min, $sec);
	
	# get seconds since the Epoch
	$epoch_secs = time();
}

### Function to compress a file using zip compression. Requires the Archive::Zip package.
### Parameters: $file_to_compress
### Return value: The name of the zip file or undef if failed to compress the file.
sub compress_file {
	print "Compressing file\n" if ($debug);
	my $file_to_compress = shift(@_);
	my $zip_file = $file_to_compress . '.zip';
	my $zip = Archive::Zip->new();
	my $file_member = $zip->addFile( $file_to_compress );
	# Save the Zip file
	unless ( $zip->writeToFileNamed($zip_file) == AZ_OK ) {
		return undef;
	}
	# Delete the uncompressed file
	unlink $file_to_compress;
	return $zip_file;
}

### Function to delete a file.
### Parameters: $file_to_delete
### Return value: 1 if successful or undef if failed
sub delete_file {
	print "Deleting file\n" if ($debug);
	my $file_to_delete = shift(@_);
	my $result = unlink $file_to_delete;
	if ( $result ) {
		return 1;
	}
	else {
		return undef;
	}
}

### Function to list files in a directory older than specified number of days.
### Requires the get_date_stamp subroutine to exist that updates the $epoch_secs var.
### See above for get_date_stamp subroutine.
### Parameters: $dir, $extension, $days
### Return value: @files
sub list_files_older_than_days {
	my $dir = shift(@_);
	my $extension = shift(@_);
	my $days = shift(@_);
	my $files;
	
	$dir = $dir . '/*';
	my @filelist = glob $dir;

	# Get a list of files in the $dir with a modification date older than $days
	my $days_in_secs = $days * 24 * 60 * 60;
	foreach my $fullpathtofile (@filelist) {
		my($filename, $directories, $suffix) = fileparse($fullpathtofile, $extension);
		if ($suffix eq $extension) {
			#print "$filename\t$directories\t$suffix\n";
			# If the file was modified more than $days days ago (using the number of seconds since 1 Jan 1970)
			my $filestat = stat($fullpathtofile);
			if ($filestat->mtime < ($epoch_secs - $days_in_secs)) {
				# push it onto a list
				push(@$files, $fullpathtofile);
			}
		}
	}
	# return that list
	return $files;
}

### Function to send an email via SMTP
### Parameters: $server, $sender, $to, $subject, @message
### Return value: None
sub send_email {
	print "Sending email\n" if ($debug);
	my $server = shift(@_);
	my $sender = shift(@_);
	my $to = shift(@_);
	my $subject = shift(@_);
	my @email_body = @_;
	my $smtp = Net::SMTP->new( $server,
				 Hello => $domainname,
				 Timeout => 10,
				 Debug =>$debug,
				 ); # open email connection
	if (! defined($smtp)){
		return undef;
	}
	$smtp->mail( $sender );	# sender
	$smtp->to( $to ); # recipients
	my $response = $smtp->data( "From: $sender\n",
				    "To: $to\n",
				    "Subject: $subject\n",
				    "\n",
				    @email_body); # open data stream
	$smtp->quit();
	return $response;
}
